PHONE(B,  "B",  "b",  "b",  "Back,BuBBle,joB"),
PHONE(CH, "CH", "tS", "tS", "CHurCH,maTCH,naTUre"),
PHONE(D,  "D",  "d",  "d",  "Day,laDDer,oDD"),
PHONE(DH, "DH", "D",  "D",  "THis,oTHer,smooTH"),
PHONE(F,  "F",  "f",  "f",  "Fat,coFFee,rouGH,PHysics"),
PHONE(G,  "G",  "g",  "g",  "Get,GiGGle,GHost"),
PHONE(HH, "HH", "h",  "h",  "Hot,WHole,beHind"),
PHONE(JH, "JH", "dZ", "dZ", "JuDGE,aGe,solDIer"),
PHONE(K,  "K",  "k",  "k",  "Key,CoCK,sCHool"),
PHONE(L,  "L",  "l",  "l",  "Light,vaLLey,feeL"),
PHONE(M,  "M",  "m",  "m",  "More,haMMer,suM"),
PHONE(N,  "N",  "n",  "n",  "Nice,KNow,fuNNy,suN"),
PHONE(NG, "NG", "N",  "N",  "riNG,loNG,thaNks,suNG"),
PHONE(P,  "P",  "p",  "p",  "Pen,coPy,haPPen"),
PHONE(R,  "R",  "r",  "r",  "Right,soRRy,aRRange"),
PHONE(S,  "S",  "s",  "s",  "Soon,CeaSe,SiSter"),
PHONE(SH, "SH", "S",  "S",  "SHip,Sure,staTIon"),
PHONE(T,  "T",  "t",  "t",  "Tea,TighT,buTTon"),
PHONE(TH, "TH", "T",  "T",  "THing,auTHor,paTH"),
PHONE(TS, "TS", "ts", "ts", "TSai"),
PHONE(V,  "V",  "v",  "v",  "View,heaVy,moVe"),
PHONE(W,  "W",  "w",  "w",  "Wet,One,When,qUeen"),
PHONE(Y,  "Y",  "j",  "j",  "Yet,Use,bEAuty"),
PHONE(Z,  "Z",  "z",  "z",  "Zero,Zone,roSeS,buZZ"),
PHONE(ZH, "ZH", "Z",  "Z",  "pleaSure,viSIon"),
PHONE(AX, "AX", "@",  "@",  "About,commA,commOn"),
PHONE(AA, "AA", "A:",  "A:",  "stARt,fAther"),
PHONE(AA1,"AA1","A:",  "A:",  "stARt,fAther"),
PHONE(AA2,"AA2","A:",  "A:",  "stARt,fAther"),
PHONE(AE, "AE", "{",  "{",  "trAp,bAd"),
PHONE(AE1,"AE1","'{", "'{", "trAp,bAd"),
PHONE(AE2,"AE2",",{", ",{", "trAp,bAd"),
PHONE(AH, "AH", "@",  "V",  "strUt,bUd,lOve"),
PHONE(AH1,"AH1","'V", "'V","strUt,bUd,lOve"),
PHONE(AH2,"AH2",",V", ",V","strUt,bUd,lOve"),
PHONE(AO, "AO", "O:",  "O:",  "thOUGHt,lAW"),
PHONE(AO1,"AO1","'O:", "'O:", "thOUGHt,lAW"),
PHONE(AO2,"AO2",",O:", ",O:", "thOUGHt,lAW"),
PHONE(AW, "AW", "aU", "aU", "mOUth,nOW"),
PHONE(AW1,"AW1","'aU","'aU","mOUth,nOW"),
PHONE(AW2,"AW2",",aU",",aU","mOUth,nOW"),
PHONE(AY, "AY", "aI", "aI", "prIce,hIGH,trY"),
PHONE(AY1,"AY1","'aI","'aI","prIce,hIGH,trY"),
PHONE(AY2,"AY2",",aI",",aI","prIce,hIGH,trY"),
PHONE(EA, "EA", "e@", "e@", "squARE,fAIR"),
PHONE(EA1,"EA1","'e@","'e@","squARE,fAIR"),
PHONE(EA2,"EA2",",e@",",e@","squARE,fAIR"),
PHONE(EH, "EH", "E",  "E",  "drEss,bEd"),
PHONE(EH1,"EH1","'E", "'E", "drEss,bEd"),
PHONE(EH2,"EH2",",E", ",E", "drEss,bEd"),
PHONE(ER, "ER", "3:",  "@`", "nURse,stIR,cOURage"),
PHONE(ER1,"ER1","'3:", "'3`","nURse,stIR,cOURage"),
PHONE(ER2,"ER2",",3:", ",3`","nURse,stIR,cOURage"),
PHONE(EY, "EY", "eI", "eI", "fAce,dAy,stEAk"),
PHONE(EY1,"EY1","'eI","'eI","fAce,dAy,stEAk"),
PHONE(EY2,"EY2",",eI",",eI","fAce,dAy,stEAk"),
PHONE(IA, "IA", "I@", "I@", "nEAR,hERE,sErious"),
PHONE(IA1,"IA1","'I@","'I@","nEAR,hERE,sErious"),
PHONE(IA2,"IA2",",I@",",I@","nEAR,hERE,sErious"),
PHONE(IH, "IH", "I",  "I",  "Intend,basIc,kIt,bId,hYmn"),
PHONE(IH1,"IH1","'I", "'I", "Intend,basIc,kIt,bId,hYmn"),
PHONE(IH2,"IH2",",I", ",I", "Intend,basIc,kIt,bId,hYmn"),
PHONE(IY, "IY", "i:",  "i:",  "happY,radIation,glorIous,flEEce,sEA,machIne"),
PHONE(IY1,"IY1","'i:", "'i:", "happY,radIation,glorIous,flEEce,sEA,machIne"),
PHONE(IY2,"IY2","%i:", ",i:", "happY,radIation,glorIous,flEEce,sEA,machIne"),
PHONE(OH, "OH", "Q",  "Q",  "lOt,Odd,wAsh"),
PHONE(OH1,"OH1","'Q", "'Q", "lOt,Odd,wAsh"),
PHONE(OH2,"OH2",",Q", ",Q", "lOt,Odd,wAsh"),
PHONE(OW, "OW", "@U", "oU", "goat,shOW,nO"),
PHONE(OW1,"OW1","'@U","'oU","goat,shOW,nO"),
PHONE(OW2,"OW2",",@U",",oU","goat,shOW,nO"),
PHONE(OY, "OY", "OI", "OI", "chOIce,bOY"),
PHONE(OY1,"OY1","'OI","'OI","chOIce,bOY"),
PHONE(OY2,"OY2",",OI",",OI","chOIce,bOY"),
PHONE(UA, "UA", "U@", "u",  "inflUence,sitUation,annUal,cURE,pOOR,jUry"),
PHONE(UA1,"UA1","'U@","'u", "inflUence,sitUation,annUal,cURE,pOOR,jUry"),
PHONE(UA2,"UA2",",U@",",u", "inflUence,sitUation,annUal,cURE,pOOR,jUry"),
PHONE(UH, "UH", "U",  "U",  "fOOt,gOOd,pUt,stimUlus,edUcate"),
PHONE(UH1,"UH1","'U", "'U", "fOOt,gOOd,pUt,stimUlus,edUcate"),
PHONE(UH2,"UH2",",U", ",U", "fOOt,gOOd,pUt,stimUlus,edUcate"),
PHONE(UW, "UW", "u:",  "u:",  "gOOse,twO,blUE"),
PHONE(UW1,"UW1","'u:", "'u:", "gOOse,twO,blUE"),
PHONE(UW2,"UW2",",u:", ",u:", "gOOse,twO,blUE"),
