#ifndef __PHFEAT_H
#define __PHFEAT_H

#define unr 0x00000000
#define rnd 0x00000001
#define low 0x00000002
#define lmd 0x00000004
#define mdl 0x00000008
#define smh 0x00000010
#define lax 0x00000020
#define hgh 0x00000040
#define bck 0x00000080
#define cnt 0x00000100
#define fnt 0x00000200
#define vwl 0x00000400
#define dip 0x00000800
#define blb 0x00001000
#define stl 0x00002000
#define vls 0x00004000
#define stp 0x00008000
#define vcd 0x00010000
#define alv 0x00020000
#define vel 0x00040000
#define glt 0x00080000
#define nas 0x00100000
#define tap 0x00200000
#define apr 0x00400000
#define rzd 0x00800000
#define frc 0x01000000
#define lbd 0x02000000
#define dnt 0x04000000
#define pla 0x08000000
#define lat 0x10000000
#define gld 0x20000000
#define lbv 0x40000000
#define pal 0x80000000

#endif
